#  Copyright (C) 2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from qgis.PyQt.QtCore import QObject

from quality_result_gui.layer_mapping import LayerMapping


class QualityResultManagerSettings(QObject):
    """
    Singleton for storing common properties.
    """

    _instance: Optional["QualityResultManagerSettings"] = None
    layer_mapping: LayerMapping = LayerMapping(layer_map=None)

    @staticmethod
    def get() -> "QualityResultManagerSettings":
        if QualityResultManagerSettings._instance is None:
            QualityResultManagerSettings._instance = QualityResultManagerSettings()
        return QualityResultManagerSettings._instance

    def set_layer_mapping(self, layer_mapping: LayerMapping) -> None:
        self.layer_mapping = layer_mapping



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from typing import Any, Dict, List

from qgis.core import QgsGeometry

from quality_result_gui.api.types.quality_error import (
    QualityError,
    QualityErrorPriority,
    QualityErrorType,
)


@dataclass
class QualityErrorResponse:
    quality_results: List[QualityError] = field(init=False)
    _errors_obj: List[Dict[str, Any]]

    def __post_init__(self) -> None:
        self.quality_results = [
            QualityError(
                priority=QualityErrorPriority(error_obj["priority"]),
                feature_type=error_obj["feature_type"],
                feature_id=error_obj["feature_id"],
                error_id=error_obj["error_id"],
                unique_identifier=error_obj["unique_identifier"],
                error_type=QualityErrorType(error_obj["error_type"]),
                attribute_name=error_obj["attribute_name"],
                error_description=error_obj["error_description"],
                error_extra_info=error_obj.get("extra_info", None),
                geometry=QgsGeometry.fromWkt(error_obj["wkt_geom"]),
                is_user_processed=error_obj["is_user_processed"],
            )
            for error_obj in self._errors_obj
        ]



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import json
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

from qgis.core import QgsCoordinateReferenceSystem

from quality_result_gui.api.quality_api_client import QualityResultClient
from quality_result_gui.api.types.quality_error import QualityError
from quality_result_gui_plugin.dev_tools.response_parser import QualityErrorResponse


@dataclass
class MockQualityResultClient(QualityResultClient):
    json_file_path: Path

    def get_results(self) -> Optional[List[QualityError]]:
        """
        Retrieve latest quality errors from API

        Returns:
            None: if no results available
            List[QualityError]: if results available

        Raises:
            QualityResultClientError: if request fails
            QualityResultServerError: if check failed in backend
        """
        return QualityErrorResponse(
            json.loads(self.json_file_path.read_text())
        ).quality_results

    def get_crs(self) -> QgsCoordinateReferenceSystem:
        return QgsCoordinateReferenceSystem("EPSG:3067")



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import List, Optional

from qgis.core import QgsCoordinateReferenceSystem

from quality_result_gui.api.types.quality_error import QualityError


class QualityResultClient(ABC):
    @abstractmethod
    def get_results(self) -> Optional[List[QualityError]]:
        """
        Retrieve latest quality errors from API

        Returns:
            None: if no results available
            List[QualityError]: if results available

        Raises:
            QualityResultClientError: if request fails
            QualityResultServerError: if check failed in backend
        """

    @abstractmethod
    def get_crs(self) -> QgsCoordinateReferenceSystem:
        pass


class QualityResultClientError(Exception):
    pass


class QualityResultServerError(Exception):
    pass



#  Copyright (C) 2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Dict, Optional

from qgis.core import QgsField, QgsProject, QgsVectorLayer


@dataclass
class LayerMapping:
    layer_map: Optional[Dict[str, str]]

    def get_layer_alias(self, feature_type: str) -> str:
        layer = self._get_layer(feature_type)
        if layer is None:
            return feature_type
        return layer.name()

    def get_field_alias(self, feature_type: str, field_name: str) -> str:
        def predicate(field: QgsField) -> bool:
            return field.name() == field_name

        layer = self._get_layer(feature_type)
        if layer is None:
            return field_name
        attribute = next(filter(predicate, layer.fields()), None)
        if attribute is None:
            return field_name
        return attribute.displayName()

    def _get_layer(self, feature_type: str) -> Optional[QgsVectorLayer]:
        if not self.layer_map:
            return None
        layer_id = self.layer_map.get(feature_type)
        if not layer_id:
            return None
        return next(
            (
                layer
                for layer in QgsProject.instance().mapLayers().values()
                if layer.id() == layer_id
            ),
            None,
        )
